@@ -24,8 +24,9 @@ class Controller(clihelper.Controller):
         :rtype: rejected.mcp.MasterControlProgram
 
         """
-        return mcp.MasterControlProgram(self._get_application_config(),
-                                        consumer=self._options.consumer)
+        return mcp.MasterControlProgram(self._config,
+                                        consumer=self._options.consumer,
+                                        profile=self._options.profile)
 
     def _prepend_python_path(self, path):  #pragma: no cover
         """Add the specified value to the python path.
@@ -44,7 +45,7 @@ class Controller(clihelper.Controller):
 
     def _shutdown(self):
         """Shutdown the MCP and child processes cleanly"""
-        self._set_state(self._STATE_SHUTTING_DOWN)
+        self.set_state(self.STATE_STOPPING)
         signal.setitimer(signal.ITIMER_PROF, 0, 0)
         self._mcp.stop_processes()
         if self._mcp.is_running:
@@ -61,13 +62,19 @@ class Controller(clihelper.Controller):
             LOGGER.info('MCP exited cleanly')
 
         # Change our state
-        self._shutdown_complete()
+        self._stopped()
 
     def run(self):
         """Run the rejected Application"""
         self._setup()
         self._mcp = self._master_control_program()
-        self._mcp.run()
+        try:
+            self._mcp.run()
+        except KeyboardInterrupt:
+            LOGGER.info('Caught CTRL-C, shutting down')
+            clihelper.setup_logging(self._debug)
+        if self.is_running:
+            self._shutdown()
 
 
 def _cli_options(parser):
@@ -76,6 +83,12 @@ def _cli_options(parser):
     :param optparse.OptionParser parser: The option parser to add options to
 
     """
+    parser.add_option('-P', '--profile',
+                      action='store',
+                      default=None,
+                      dest='profile',
+                      help='Profile the consumer modules, specifying '
+                           'the output directory.')
     parser.add_option('-o', '--only',
                       action='store',
                       default=None,
