@@ -5,6 +5,7 @@ Master Control Program
 import logging
 import multiprocessing
 import os
+import psutil
 import Queue
 import signal
 import sys
@@ -27,11 +28,13 @@ class MasterControlProgram(state.State):
     _POLL_RESULTS_INTERVAL = 3.0
     _SHUTDOWN_WAIT = 1
 
-    def __init__(self, config, consumer=None):
+    def __init__(self, config, consumer=None, profile=None):
         """Initialize the Master Control Program
 
         :param dict config: The full content from the YAML config file
         :param str consumer: If specified, only run processes for this consumer
+        :param str profile: Optional profile output directory to
+                            enable profiling
 
         """
         self._set_process_name()
@@ -45,6 +48,7 @@ class MasterControlProgram(state.State):
         self._last_poll_results = dict()
         self._poll_data = {'time': 0, 'processes': list()}
         self._poll_timer = None
+        self._profile = profile
         self._results_timer = None
         self._stats = dict()
         self._stats_queue = multiprocessing.Queue()
@@ -64,19 +68,58 @@ class MasterControlProgram(state.State):
         :rtype: list
 
         """
-        dead_processes = list()
-        active_processes = multiprocessing.active_children()
-        for process in active_processes:
-            if not process.is_alive():
-                dead_processes.append(process)
+        active_processes, dead_processes = list(), list()
+        for consumer in self._consumers:
+            for name in self._consumers[consumer]['processes']:
+                child = self.get_consumer_process(consumer, name)
+                try:
+                    process = psutil.Process(child.pid)
+                except psutil.NoSuchProcess:
+                    dead_processes.append((consumer, name))
+                    continue
+
+                if self.is_a_dead_or_zombie_process(process):
+                    dead_processes.append((consumer, name))
+                else:
+                    active_processes.append(child)
+
         if dead_processes:
-            LOGGER.debug('Found %i dead processes to remove',
-                         len(dead_processes))
+            LOGGER.info('Removing %i dead process(es)', len(dead_processes))
             for process in dead_processes:
-                self._remove_process(process.name)
-                active_processes.remove(process)
+                self.remove_consumer_process(*process)
         return active_processes
 
+    def is_a_dead_or_zombie_process(self, process):
+        try:
+            if process.status in [psutil.STATUS_DEAD, psutil.STATUS_ZOMBIE]:
+                try:
+                    LOGGER.info('Found dead or zombie process with '
+                                '%s fds',
+                                process.get_num_fds())
+                except psutil.AccessDenied as error:
+                    LOGGER.error('Found dead or zombie process, '
+                                 'could not get fd count: %s', error)
+                    return True
+        except psutil.NoSuchProcess:
+            LOGGER.info('Found dead or zombie process')
+            return True
+
+        return False
+
+    def get_consumer_process(self, consumer, name):
+        return self._consumers[consumer]['processes'].get(name)
+
+    def remove_consumer_process(self, consumer, name):
+        for conn in self._consumers[consumer]['connections']:
+            if name in self._consumers[consumer]['connections'][conn]:
+                self._consumers[consumer]['connections'][conn].remove(name)
+        if name in self._consumers[consumer]['processes']:
+            try:
+                self._consumers[consumer]['processes'][name].terminate()
+            except OSError:
+                pass
+            del self._consumers[consumer]['processes'][name]
+
     def _calculate_stats(self, data):
         """Calculate the stats data for our process level data.
 
@@ -129,7 +172,6 @@ class MasterControlProgram(state.State):
         LOGGER.debug('Checking minimum consumer process levels')
         for name in self._consumers:
             for connection in self._consumers[name]['connections']:
-
                 processes_needed = self._process_spawn_qty(name, connection)
                 if processes_needed:
                     LOGGER.debug('Need to spawn %i processes for %s on %s',
@@ -227,15 +269,11 @@ class MasterControlProgram(state.State):
 
         """
         LOGGER.critical('Max shutdown exceeded, forcibly exiting')
-
-        for process in multiprocessing.active_children():
-            process.terminate()
-        LOGGER.info('All processes should have terminated')
-        time.sleep(0.5)
-        while multiprocessing.active_children():
-            for process in multiprocessing.active_children():
-                LOGGER.warning('Killing %s (%s)', process.name, process.pid)
-                os.kill(int(process.pid), signal.SIGKILL)
+        while self._active_processes:
+            for process in self._active_processes:
+                if int(process.pid) != int(os.getpid()):
+                    LOGGER.warning('Killing %s (%s)', process.name, process.pid)
+                    os.kill(int(process.pid), signal.SIGKILL)
             time.sleep(1)
 
         return self._set_state(self.STATE_STOPPED)
@@ -284,9 +322,11 @@ class MasterControlProgram(state.State):
                                   self._new_process_number(consumer_name))
         LOGGER.debug('Creating a new process for %s: %s',
                      connection_name, process_name)
-        kwargs = {'config': self._config,
+        kwargs = {'config': self._config['Application'],
                   'connection_name': connection_name,
                   'consumer_name': consumer_name,
+                  'profile': self._profile,
+                  'daemon': False,
                   'stats_queue': self._stats_queue}
         return process_name, process.Process(name=process_name, kwargs=kwargs)
 
@@ -307,14 +347,6 @@ class MasterControlProgram(state.State):
         processing, note it with a warning.
 
         """
-        # Check to see if we have outstanding things to poll
-        #if self._poll_data['processes']:
-        #    LOGGER.warn('Poll interval failure for consumer(s): %r',
-        #                 ', '.join(self._poll_data['processes']))
-
-        # Keep track of running consumers
-        dead_processes = list()
-
         # Start our data collection dict
         self._poll_data = {'timestamp': time.time(),
                            'processes': list()}
@@ -326,26 +358,14 @@ class MasterControlProgram(state.State):
                 LOGGER.debug('Matched current process in active_processes')
                 continue
 
-            if not process.is_alive():
-                LOGGER.critical('Found dead consumer %s', process.name)
-                dead_processes.append(process.name)
-            else:
-                LOGGER.debug('Asking %s for stats', process.name)
-                #self._poll_data['processes'].append(process.name)
-                os.kill(process.pid, signal.SIGPROF)
-
-        # Remove the objects if we have them
-        for process_name in dead_processes:
-            self._remove_process(process_name)
-
-        # Check if we need to start more processes
-        self._check_consumer_process_counts()
-
         # If we don't have any active consumers, shutdown
         if not self.total_process_count:
             LOGGER.debug('Did not find any active consumers in poll')
             return self._set_state(self.STATE_STOPPED)
 
+        # Check if we need to start more processes
+        self._check_consumer_process_counts()
+
         # Check to see if any consumers reported back and start timer if not
         #self._poll_results_check()
 
@@ -358,8 +378,8 @@ class MasterControlProgram(state.State):
         :rtype: bool
 
         """
-        return  (time.time() -
-                 self._poll_data['timestamp']) >= self._poll_interval
+        return (time.time() -
+                self._poll_data['timestamp']) >= self._poll_interval
 
     def _poll_results_check(self):
         """Check the polling results by checking to see if the stats queue is
@@ -432,37 +452,6 @@ class MasterControlProgram(state.State):
         return self._consumers[name]['min'] - self._process_count(name,
                                                                   connection)
 
-    def _remove_process(self, process):
-        """Remove the specified consumer process
-
-        :param str process: The process name to remove
-
-        """
-        for consumer_name in self._consumers:
-            if process in self._consumers[consumer_name]['processes']:
-                del self._consumers[consumer_name]['processes'][process]
-                self._remove_process_from_list(consumer_name, process)
-                LOGGER.debug('Removed %s from %s\'s process list',
-                             process, consumer_name)
-                return
-        LOGGER.warning('Could not find process %s to remove', process)
-
-    def _remove_process_from_list(self, name, process):
-        """Remove the process from the consumer connections list.
-
-        :param str name: The consumer name
-        :param str process: The process name
-
-        """
-        for conn in self._consumers[name]['connections']:
-            if process in self._consumers[name]['connections'][conn]:
-                idx = self._consumers[name]['connections'][conn].index(process)
-                self._consumers[name]['connections'][conn].pop(idx)
-                LOGGER.debug('Process %s removed from %s connections',
-                             process, name)
-                return
-        LOGGER.warning('Could not find %s in %s\'s process list', process, name)
-
     def _set_process_name(self):
         """Set the process name for the top level process so that it shows up
         in logs in a more trackable fasion.
@@ -478,6 +467,7 @@ class MasterControlProgram(state.State):
     def _start_poll_results_timer(self):
         """Start the poll results timer to see if there are results from the
         last poll yet.
+
         """
         self._results_timer = self._start_timer(self._results_timer,
                                                 'poll_results_timer',
@@ -499,7 +489,6 @@ class MasterControlProgram(state.State):
         LOGGER.info('Spawning new consumer process for %s to %s',
                     name, connection)
 
-        # Create the new consumer process
         process_name, process = self._new_process(name, connection)
 
         # Append the process to the consumer process list
@@ -508,6 +497,7 @@ class MasterControlProgram(state.State):
 
         # Start the process
         process.start()
+        #process.join(0.1)
 
     def _start_processes(self, name, connection, quantity):
         """Start the specified quantity of consumer processes for the given
@@ -524,11 +514,11 @@ class MasterControlProgram(state.State):
     def _start_timer(self, timer, name, duration, callback):
         """Start a timer for the given object, name, duration and callback.
 
-        :param threading.Timer timer: The previous timer instance
+        :param threading._Timer timer: The previous timer instance
         :param str name: The timer name
         :param int or float duration: The timer duration
         :param method callback: The method to call when timer fires
-        :rtype: threading.Timer
+        :rtype: threading._Timer
 
         """
         if timer and timer.is_alive():
@@ -547,10 +537,10 @@ class MasterControlProgram(state.State):
         minimal amount of processes, setting up the runtime data as well.
 
         """
-        for name in self._config['Consumers']:
+        for name in self._config['Application']['Consumers']:
 
             # Hold the config as a shortcut
-            config = self._config['Consumers'][name]
+            config = self._config['Application']['Consumers'][name]
 
             # If queue is not configured, report the error but skip processes
             if 'queue' not in config:
@@ -566,19 +556,6 @@ class MasterControlProgram(state.State):
                 self._start_processes(name, connection,
                                       self._consumers[name]['min'])
 
-    def _stop_process(self, process):
-        """Stop the specified process
-
-        :param multiprocessing.Process process: The process to stop
-
-        """
-        if not process.is_alive():
-            return
-
-        LOGGER.debug('Sending signal to %s (%i) to stop',
-                     process.name, process.pid)
-        os.kill(int(process.pid), signal.SIGABRT)
-
     def _stop_timers(self):
         """Stop all the active timeouts."""
         if self._poll_timer and self._poll_timer.is_alive():
@@ -595,21 +572,20 @@ class MasterControlProgram(state.State):
         LOGGER.debug('Stopping consumer processes')
         self._stop_timers()
 
-        active_processes = multiprocessing.active_children()
+        signal.signal(signal.SIGCHLD, signal.SIG_IGN)
+        signal.signal(signal.SIGALRM, signal.SIG_IGN)
+        signal.signal(signal.SIGPROF, signal.SIG_IGN)
 
-        # Stop if we have no running consumers
-        if not active_processes:
-            LOGGER.info('All consumer processes have stopped')
-            return self._set_state(self.STATE_STOPPED)
-
-        # Iterate through all of the bindings and try and shutdown processes
-        for process in active_processes:
-            self._stop_process(process)
+        # Send SIGTERM
+        for process in multiprocessing.active_children():
+            process.terminate()
 
+        # Wait for them to finish up to MAX_SHUTDOWN_WAIT
         iterations = 0
-        while multiprocessing.active_children():
+        processes = self.total_process_count
+        while processes:
             LOGGER.debug('Waiting on %i active processes to shut down',
-                         self.total_process_count)
+                         processes)
             try:
                 time.sleep(1)
             except KeyboardInterrupt:
@@ -619,11 +595,10 @@ class MasterControlProgram(state.State):
                 return
 
             iterations += 1
-
-            # If the shutdown process waited long enough, kill the consumers
             if iterations == self._MAX_SHUTDOWN_WAIT:
                 self._kill_processes()
                 break
+            processes = self.total_process_count
 
         LOGGER.debug('All consumer processes stopped')
         self._set_state(self.STATE_STOPPED)
@@ -637,19 +612,22 @@ class MasterControlProgram(state.State):
         self._set_state(self.STATE_ACTIVE)
 
         # Get the consumer section from the config
-        if 'Consumers' not in self._config:
+        if 'Consumers' not in self._config['Application']:
             LOGGER.error('Missing Consumers section of configuration, '
-                         'aborting: %r', self._config)
+                         'aborting: %r', self._config['Application'])
             return self._set_state(self.STATE_STOPPED)
 
         # Strip consumers if a consumer is specified
         if self._consumer:
-            consumers = self._config['Consumers'].keys()
+            consumers = self._config['Application']['Consumers'].keys()
             for consumer in consumers:
                 if consumer != self._consumer:
                     LOGGER.debug('Removing %s for %s only processing',
                                  consumer, self._consumer)
-                    del self._config['Consumers'][consumer]
+                    del self._config['Application']['Consumers'][consumer]
+
+        # Setup the SIGCHLD handler
+        #self.setup_signal_handler()
 
         # Setup consumers and start the processes
         self._setup_consumers()
@@ -657,9 +635,9 @@ class MasterControlProgram(state.State):
         # Kick off the poll timer
         self._start_poll_timer()
 
-        # Loop for the lifetime of the app, sleeping 0.2 seconds at a time
-        while self.is_running:
-            time.sleep(0.2)
+        # Loop for the lifetime of the app, pausing for a signal to pop up
+        while self.is_running and self.total_process_count:
+            signal.pause()
 
         # Note we're exiting run
         LOGGER.info('Exiting Master Control Program')
@@ -672,3 +650,14 @@ class MasterControlProgram(state.State):
 
         """
         return len(self._active_processes)
+
+    def on_sigalrt(self, signum, frame):
+        LOGGER.debug('Received %s', signum)
+        if self.is_running and self._active_processes:
+            self._poll()
+        self.setup_signal_handler()
+
+    def setup_signal_handler(self):
+        LOGGER.debug('Setting up signal handler for SIGALRT')
+        signal.signal(signal.SIGALRT, self.on_sigalrt)
+        signal.siginterrupt(signal.SIGALRT, False)
