@@ -3,18 +3,25 @@ connection state and collects stats about the consuming process.
 
 """
 from pika import exceptions
+import importlib
 import logging
 import math
 import multiprocessing
 import os
 import pika
+try:
+    import cProfile as profile
+except ImportError:
+    import profile
 from pika.adapters import tornado_connection
 import signal
+import socket
 import sys
 import time
 from tornado import ioloop
 import traceback
 
+
 from rejected import __version__
 from rejected import consumer
 from rejected import data
@@ -23,30 +30,25 @@ from rejected import state
 LOGGER = logging.getLogger(__name__)
 
 
-def import_namespaced_class(namespaced_class):
+def import_consumer(consumer):
     """Pass in a string in the format of foo.Bar, foo.bar.Baz, foo.bar.baz.Qux
     and it will return a handle to the class
 
-    :param str namespaced_class: The namespaced class
+    :param str consumer: The consumer class in module.Consumer format
     :return: tuple(Class, str)
 
     """
-    LOGGER.debug('Importing %s', namespaced_class)
-    # Split up our string containing the import and class
-    parts = namespaced_class.split('.')
 
-    # Build our strings for the import name and the class name
+    parts = consumer.split('.')
     import_name = '.'.join(parts[0:-1])
-    class_name = parts[-1]
-
-    import_handle = __import__(import_name, fromlist=class_name)
+    import_handle = importlib.import_module(import_name)
     if hasattr(import_handle, '__version__'):
         version = import_handle.__version__
     else:
         version = None
 
     # Return the class handle
-    return getattr(import_handle, class_name), version
+    return getattr(import_handle, parts[-1]), version
 
 
 class Process(multiprocessing.Process, state.State):
@@ -73,7 +75,7 @@ class Process(multiprocessing.Process, state.State):
     TIME_WAITED = 'idle_time'
     UNHANDLED_EXCEPTIONS = 'unhandled_exceptions'
 
-    _HBINTERVAL = 30
+    _HBINTERVAL = 120
 
     # Locations to search for newrelic ini files
     INI_DIRS = ['.', '/etc/', '/etc/newrelic']
@@ -110,9 +112,13 @@ class Process(multiprocessing.Process, state.State):
         self._message_connection_id = None
         self._max_framesize = pika.spec.FRAME_MAX_SIZE
         self._qos_prefetch = None
+        self._prepend_path = None
         self._state = self.STATE_INITIALIZING
         self._state_start = time.time()
         self._stats_queue = None
+        self._statsd = False
+        self._statsd_host = 'localhost'
+        self._statsd_port = 8125
 
         # Override ACTIVE with PROCESSING
         self._STATES[0x04] = 'Processing'
@@ -128,7 +134,7 @@ class Process(multiprocessing.Process, state.State):
             self.increment_count(self.CLOSED_ON_COMPLETE)
             return
         LOGGER.debug('Acking %s', delivery_tag)
-        self._channel.basic_ack(delivery_tag=delivery_tag)
+        self._channel.basic_ack(delivery_tag=delivery_tag, multiple=True)
         self.increment_count(self.ACKED)
 
     def add_on_channel_close_callback(self):
@@ -227,7 +233,7 @@ class Process(multiprocessing.Process, state.State):
 
         :param dict config: The Connections section of the configuration
         :param str name: The name of the connection
-        :rtype: pika.adapters.tornado_conneciton.TornadoConnection
+        :rtype: pika.adapters.tornado_connection.TornadoConnection
 
         """
         LOGGER.debug('Connecting to %s:%i:%s as %s',
@@ -242,7 +248,7 @@ class Process(multiprocessing.Process, state.State):
                                                     config[name]['pass'])
         return tornado_connection.TornadoConnection(parameters,
                                                     self.on_connection_open,
-                                                    True)
+                                                    stop_ioloop_on_close=True)
 
     def count(self, stat):
         """Return the current count quantity for a specific stat.
@@ -308,6 +314,7 @@ class Process(multiprocessing.Process, state.State):
         credentials = pika.PlainCredentials(username, password)
         return pika.ConnectionParameters(host, port, vhost, credentials,
                                          frame_max=self._max_framesize,
+                                         socket_timeout=10,
                                          heartbeat_interval=self._hbinterval)
 
     def get_consumer(self, config):
@@ -318,12 +325,11 @@ class Process(multiprocessing.Process, state.State):
         :raises: ImportError
 
         """
-        # Try and import the module
         try:
-            consumer_, version = import_namespaced_class(config['consumer'])
-        except Exception as error:
-            LOGGER.error('Error importing the consumer "%s": %s',
-                         config['consumer'], error)
+            consumer_, version = import_consumer(config['consumer'])
+        except ImportError as error:
+            LOGGER.exception('Error importing the consumer %s: %s',
+                             config['consumer'], error)
             exc_type, exc_value, exc_traceback = sys.exc_info()
             lines = traceback.extract_tb(exc_traceback)
             for line in lines:
@@ -335,22 +341,20 @@ class Process(multiprocessing.Process, state.State):
         else:
             LOGGER.info('Creating consumer %s', config['consumer'])
 
-        # If we have a config, pass it in to the constructor
+        kwargs = {}
         if 'config' in config:
-            try:
-                return consumer_(config['config'])
-            except Exception as error:
-                LOGGER.error('Error creating the consumer "%s": %s',
-                             config['consumer'], error)
-                exc_type, exc_value, exc_traceback = sys.exc_info()
-                lines = traceback.extract_tb(exc_traceback)
-                for line in lines:
-                    LOGGER.error(line)
-                return
+            kwargs['configuration'] = config['config']
 
-        # No config to pass
         try:
-            return consumer_()
+            if consumer_.WANTS_CONNECTION_CONFIG:
+                kwargs['connections'] = self._connections
+        except AttributeError:
+            pass
+
+        try:
+            return consumer_(**kwargs)
+        except TypeError:
+            return consumer_(config['config'])
         except Exception as error:
             LOGGER.error('Error creating the consumer "%s": %s',
                          config['consumer'], error)
@@ -370,6 +374,8 @@ class Process(multiprocessing.Process, state.State):
 
         """
         self._counts[counter] += value
+        if self._statsd:
+            self.send_counter_to_statsd(counter, value)
 
     def invoke_consumer(self, message):
         """Wrap the actual processor processing bits
@@ -502,10 +508,9 @@ class Process(multiprocessing.Process, state.State):
         LOGGER.info('Channel opened')
         self._channel = channel
         self.add_on_channel_close_callback()
-        self.set_state(self.STATE_IDLE)
         self.setup_channel()
 
-    def on_connection_closed(self, unused):
+    def on_connection_closed(self, unused, code, text):
         """This method is invoked by pika when the connection to RabbitMQ is
         closed unexpectedly. Since it is unexpected, we will reconnect to
         RabbitMQ if it disconnects.
@@ -513,9 +518,10 @@ class Process(multiprocessing.Process, state.State):
         :param pika.connection.Connection unused: The closed connection
 
         """
-        LOGGER.critical('Connection from RabbitMQ closed (%r)', self._state)
+        LOGGER.critical('Connection from RabbitMQ closed in state %i (%s, %s)',
+                        self.state_description, code, text)
         self._channel = None
-        if not self.is_shutting_down:
+        if not self.is_shutting_down and not self.is_waiting_to_shutdown:
             self.reconnect()
 
     def on_connection_open(self, unused):
@@ -535,6 +541,12 @@ class Process(multiprocessing.Process, state.State):
         # Set the state to shutting down if it wasn't set as that during loop
         self.set_state(self.STATE_SHUTTING_DOWN)
 
+        # Reset any signal handlers
+        signal.signal(signal.SIGABRT, signal.SIG_IGN)
+        signal.signal(signal.SIGINT, signal.SIG_IGN)
+        signal.signal(signal.SIGPROF, signal.SIG_IGN)
+        signal.signal(signal.SIGTERM, signal.SIG_IGN)
+
         # If the connection is still around, close it
         if self._connection.is_open:
             LOGGER.debug('Closing connection to RabbitMQ')
@@ -544,11 +556,14 @@ class Process(multiprocessing.Process, state.State):
         self.stop_consumer()
 
         # Stop the IOLoop
+        LOGGER.info('Stopping IOLoop')
         self._connection.ioloop.stop()
 
         # Note that shutdown is complete and set the state accordingly
-        LOGGER.info('Shutdown complete')
         self.set_state(self.STATE_STOPPED)
+        LOGGER.info('Shutdown complete')
+        os.kill(os.getppid(), signal.SIGALRM)
+        os._exit(0)
 
     def on_sigprof(self, unused_signum, unused_frame):
         """Called when SIGPROF is sent to the process, will dump the stats, in
@@ -559,6 +574,7 @@ class Process(multiprocessing.Process, state.State):
 
         """
         LOGGER.info('Currently %s: %r', self.state_description, self._counts)
+        signal.siginterrupt(signal.SIGPROF, False)
 
     def open_channel(self):
         """Open a channel on the existing open connection to RabbitMQ"""
@@ -590,8 +606,22 @@ class Process(multiprocessing.Process, state.State):
         self.increment_count(self.PROCESSED)
         if self._ack:
             self.ack_message(method.delivery_tag)
+        if self.is_waiting_to_shutdown:
+            return self.on_ready_to_stop()
         self.reset_state()
 
+    @property
+    def profile_file(self):
+        if not self._kwargs['profile']:
+            return None
+
+        if os.path.exists(self._kwargs['profile']) and \
+                os.path.isdir(self._kwargs['profile']):
+            return '%s/%s-%s.prof' % (os.path.normpath(self._kwargs['profile']),
+                                      os.getpid(),
+                                      self._kwargs['consumer_name'])
+        return None
+
     def processing_failure(self):
         """Called when message processing failure happens due to a
         ConsumerException or an unhandled exception.
@@ -697,6 +727,16 @@ class Process(multiprocessing.Process, state.State):
 
     def run(self):
         """Start the consumer"""
+        if self.profile_file:
+            LOGGER.info('Profiling to %s', self.profile_file)
+            profile.runctx('self._run()', globals(), locals(),
+                           self.profile_file)
+        else:
+            self._run()
+        LOGGER.info('Exiting %s (%i, %i)', self.name, os.getpid(), os.getppid())
+
+    def _run(self):
+        """Run method that can be profiled"""
         try:
             self.setup(self._kwargs['config'],
                        self._kwargs['connection_name'],
@@ -709,20 +749,19 @@ class Process(multiprocessing.Process, state.State):
                             consumer, error)
             return
 
-        while not self.is_waiting_to_shutdown and not self.is_stopped:
+        if not self.is_stopped:
             try:
                 self._connection.ioloop.start()
             except KeyboardInterrupt:
-                self.stop()
-                try:
-                    self._connection.ioloop.start()
-                except KeyboardInterrupt:
-                    LOGGER.warning('CTRL-C while waiting for clean shutdown')
-            if not self.is_stopped:
-                while not self._connection and self.is_connecting:
-                    time.sleep(0.1)
+                LOGGER.warning('CTRL-C while waiting for clean shutdown')
 
-        LOGGER.info('Exiting %s', self.name)
+    def send_counter_to_statsd(self, counter, value=1):
+        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
+        sock.connect((self._statsd_host, self._statsd_port))
+        key = '.'.join(['rejected', 'consumers', self._kwargs['consumer_name'],
+                        counter])
+        sock.send('%s:%i|c\n' % (key, value))
+        sock.close()
 
     def set_qos_prefetch(self, value=None):
         """Set the QOS Prefetch count for the channel.
@@ -753,15 +792,15 @@ class Process(multiprocessing.Process, state.State):
         # Use the parent object to set the state
         super(Process, self)._set_state(new_state)
 
-    def setup(self, config, connection_name, consumer_name, stats_queue):
+    def setup(self, config, connection_name,
+              consumer_name, stats_queue):
         """Initialize the consumer, setting up needed attributes and connecting
         to RabbitMQ.
 
         :param dict config: Consumer config section
-        :param str connection_name: The name of the connection
+\        :param str connection_name: The name of the connection
         :param str consumer_name: Consumer name for config
         :param multiprocessing.Queue stats_queue: The queue to append stats in
-        :raises: ImportError
 
         """
         LOGGER.info('Initializing for %s on %s', self.name, connection_name)
@@ -775,10 +814,18 @@ class Process(multiprocessing.Process, state.State):
         self._config = config['Consumers'][consumer_name]
         self._connections = config['Connections']
 
+        self._statsd = False
+        if 'statsd' in config and config['statsd'].get('enabled', False):
+            self._statsd = True
+            self._statsd_host = config['statsd'].get('host', 'localhost')
+            self._statsd_port = config['statsd'].get('port', 8125)
+
         # Setup the consumer
         self._consumer = self.get_consumer(self._config)
         if not self._consumer:
-            raise ImportError('Could not import and start processor')
+            LOGGER.critical('Could not import and start processor')
+            self.set_state(self.STATE_STOPPED)
+            os._exit(1)
 
         # Set the routing information
         self._queue_name = self._config['queue']
@@ -815,6 +862,8 @@ class Process(multiprocessing.Process, state.State):
         consumer object.
 
         """
+        self.set_state(self.STATE_IDLE)
+
         # Set the channel in the consumer
         try:
             self._consumer.set_channel(self._channel)
@@ -830,14 +879,17 @@ class Process(multiprocessing.Process, state.State):
                                     consumer_tag=self.name)
 
     def setup_signal_handlers(self):
-        """Setup the stats and stop signal handlers. Use SIGABRT instead of
-        SIGTERM due to the multiprocessing's behavior with SIGTERM.
-
+        """Setup the stats and stop signal handlers.
         """
-        signal.signal(signal.SIGABRT, self.stop)
+        signal.signal(signal.SIGABRT, signal.SIG_IGN)
+        signal.signal(signal.SIGCHLD, signal.SIG_IGN)
+        signal.signal(signal.SIGINT, signal.SIG_IGN)
         signal.signal(signal.SIGPROF, self.on_sigprof)
+        signal.signal(signal.SIGTERM, self.stop)
         signal.siginterrupt(signal.SIGABRT, False)
+        signal.siginterrupt(signal.SIGINT, False)
         signal.siginterrupt(signal.SIGPROF, False)
+        signal.siginterrupt(signal.SIGTERM, False)
 
     def start_message_processing(self):
         """Keep track of the connection in case RabbitMQ disconnects while the
@@ -851,26 +903,26 @@ class Process(multiprocessing.Process, state.State):
         our state.
 
         """
-        LOGGER.debug('Stop called in state: %s', self.state_description)
+        LOGGER.warning('Stop called in state: %s', self.state_description)
         if self.is_stopped:
-            LOGGER.debug('Stop requested but consumer is already stopped')
+            LOGGER.warning('Stop requested but consumer is already stopped')
             return
         elif self.is_shutting_down:
-            LOGGER.debug('Stop requested but consumer is already shutting down')
+            LOGGER.warning('Stop requested but consumer is already shutting down')
             return
         elif self.is_waiting_to_shutdown:
-            LOGGER.debug('Stop requested but already waiting to shut down')
+            LOGGER.warning('Stop requested but already waiting to shut down')
             return
 
         # Stop consuming
-        LOGGER.info('Shutting down')
         self.cancel_consumer_with_rabbitmq()
 
         # Wait until the consumer has finished processing to shutdown
         if self.is_processing:
+            LOGGER.info('Waiting for consumer to finish processing')
             self.set_state(self.STATE_STOP_REQUESTED)
-            if signum == signal.SIGABRT:
-                signal.siginterrupt(signal.SIGABRT, False)
+            if signum == signal.SIGTERM:
+                signal.siginterrupt(signal.SIGTERM, False)
             return
 
         self.on_ready_to_stop()
